import glfw
from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLUT import GLUT_BITMAP_HELVETICA_18
import time
import random
import inspect
import py_compile
import re
import shutil
import os

class Key:
    SPACE = glfw.KEY_SPACE
    UP = glfw.KEY_UP
    W = glfw.KEY_W

class Process:
    def __init__(self, engine):
        self.engine = engine
        self.frame_callback = None

    def Frame(self, callback):
        self.frame_callback = callback

    def run(self):
        if not glfw.init():
            return
        
        # Initialize GLUT
        glutInit()

        window = glfw.create_window(self.engine.window_size[0], self.engine.window_size[1], self.engine.window_title, None, None)

        if not window:
            glfw.terminate()
            return

        glfw.make_context_current(window)
        glOrtho(0, self.engine.window_size[0], 0, self.engine.window_size[1], -1, 1)
        glClearColor(self.engine.window_background_color[0] / 255.0,
                     self.engine.window_background_color[1] / 255.0,
                     self.engine.window_background_color[2] / 255.0, 1.0)

        last_time = time.time()

        while not glfw.window_should_close(window):
            current_time = time.time()
            delta_time = current_time - last_time
            last_time = current_time

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

            self.engine.handle_input(window)
            self.engine.update_bird(delta_time)
            self.engine.update_pipes(delta_time)
            self.engine.check_collision()
            self.engine.draw_bird()
            self.engine.draw_pipes()
            self.engine.draw_score()

            if self.frame_callback:
                self.frame_callback()

            glfw.swap_buffers(window)
            glfw.poll_events()
            time.sleep(1.0 / self.engine.window_fps)  # Control the frame rate

        glfw.terminate()

class Engine:
    def __init__(self):
        self.window_size = (1000, 1000)  # Default window size
        self.window_title = "Flappy Engine"  # Default window title
        self.jump_key = Key.SPACE  # Default jump key
        self.window_fps = 60  # Default FPS
        self.window_background_color = (0, 0, 100)  # Default background color
        self.bird_gravity = 10  # Default bird gravity
        self.bird_jump_force = 1000  # Default jump force
        self.bird_jump_cooldown = 0.5  # Default jump cooldown
        self.bird_color = (255, 255, 0)  # Default bird color
        self.bird_position = [self.window_size[0] / 4, self.window_size[1] / 2]
        self.bird_velocity = 0
        self.last_jump_time = 0
        self.bird_size = (50, 50)  # Default bird size
        self.pipe_speed = 1000  # Default pipe speed
        self.pipe_gap = 150  # Default pipe gap
        self.pipe_frequency = 0.1  # Default pipe frequency
        self.pipe_color = (0, 255, 0)  # Default pipe color
        self.pipe_width = 100  # Default pipe width
        self.pipes = []
        self.last_pipe_time = 0
        self.score = 0  # Initialize score
        self.score_color = (255, 0, 0)  # Default score color (red)

    def SetWindowSize(self, width, height):
        self.window_size = (width, height)

    def SetWindowTitle(self, title):
        self.window_title = title

    def SetJumpKey(self, key):
        self.jump_key = key

    def SetWindowFPS(self, fps):
        self.window_fps = fps

    def SetWindowBackgroundColor(self, r, g, b):
        self.window_background_color = (r, g, b)

    def SetBirdGravity(self, gravity):
        self.bird_gravity = gravity

    def SetBirdJumpForce(self, force):
        self.bird_jump_force = force

    def SetBirdJumpCooldown(self, cooldown):
        self.bird_jump_cooldown = cooldown

    def SetBirdColor(self, r, g, b):
        self.bird_color = (r, g, b)

    def SetBirdSize(self, width, height):
        self.bird_size = (width, height)

    def SetPipeSpeed(self, speed):
        self.pipe_speed = speed

    def SetPipeGap(self, gap):
        self.pipe_gap = gap

    def SetPipeFrequency(self, frequency):
        self.pipe_frequency = frequency

    def SetPipeColor(self, r, g, b):
        self.pipe_color = (r, g, b)

    def SetPipeWidth(self, width):
        self.pipe_width = width

    def SetScoreColor(self, r, g, b):
        self.score_color = (r, g, b)

    def draw_bird(self):
        glColor3ub(*self.bird_color)
        glBegin(GL_QUADS)
        glVertex2f(self.bird_position[0] - self.bird_size[0] / 2, self.bird_position[1] - self.bird_size[1] / 2)
        glVertex2f(self.bird_position[0] + self.bird_size[0] / 2, self.bird_position[1] - self.bird_size[1] / 2)
        glVertex2f(self.bird_position[0] + self.bird_size[0] / 2, self.bird_position[1] + self.bird_size[1] / 2)
        glVertex2f(self.bird_position[0] - self.bird_size[0] / 2, self.bird_position[1] + self.bird_size[1] / 2)
        glEnd()

    def draw_pipes(self):
        glColor3ub(*self.pipe_color)
        for pipe in self.pipes:
            # Draw top pipe
            glBegin(GL_QUADS)
            glVertex2f(pipe['x'], pipe['top'])
            glVertex2f(pipe['x'] + self.pipe_width, pipe['top'])
            glVertex2f(pipe['x'] + self.pipe_width, self.window_size[1])
            glVertex2f(pipe['x'], self.window_size[1])
            glEnd()

            # Draw bottom pipe
            glBegin(GL_QUADS)
            glVertex2f(pipe['x'], 0)
            glVertex2f(pipe['x'] + self.pipe_width, 0)
            glVertex2f(pipe['x'] + self.pipe_width, pipe['bottom'])
            glVertex2f(pipe['x'], pipe['bottom'])
            glEnd()

    def update_bird(self, delta_time):
        self.bird_velocity -= self.bird_gravity
        self.bird_position[1] += self.bird_velocity * delta_time * 0.1

        if self.bird_position[1] < 0 or self.bird_position[1] > self.window_size[1]:
            self.reset_game()

    def update_pipes(self, delta_time):
        current_time = time.time()
        if current_time - self.last_pipe_time > 1.0 / self.pipe_frequency:
            self.last_pipe_time = current_time
            pipe_height = random.randint(100, self.window_size[1] - 100)
            self.pipes.append({
                'x': self.window_size[0],
                'top': pipe_height + self.pipe_gap / 2,
                'bottom': pipe_height - self.pipe_gap / 2
            })

        for pipe in self.pipes:
            pipe['x'] -= self.pipe_speed * delta_time * 0.1

            # Check if the bird has passed the pipe
            if not pipe.get('scored') and self.bird_position[0] > pipe['x'] + self.pipe_width:
                self.score += 1
                pipe['scored'] = True  # Mark this pipe as scored

        self.pipes = [pipe for pipe in self.pipes if pipe['x'] + self.pipe_width > 0]

    def check_collision(self):
        for pipe in self.pipes:
            if (self.bird_position[0] + self.bird_size[0] / 2 > pipe['x'] and
                self.bird_position[0] - self.bird_size[0] / 2 < pipe['x'] + self.pipe_width):
                if (self.bird_position[1] + self.bird_size[1] / 2 > pipe['top'] or
                    self.bird_position[1] - self.bird_size[1] / 2 < pipe['bottom']):
                    self.reset_game()

    def reset_game(self):
        self.bird_position = [self.window_size[0] / 4, self.window_size[1] / 2]
        self.bird_velocity = 0
        self.last_jump_time = 0
        self.pipes = []
        self.score = 0

    def handle_input(self, window):
        if glfw.get_key(window, self.jump_key) == glfw.PRESS:
            current_time = time.time()
            if current_time - self.last_jump_time > self.bird_jump_cooldown:
                self.bird_velocity = self.bird_jump_force
                self.last_jump_time = current_time

    def draw_score(self):
        glColor3f(*[c / 255.0 for c in self.score_color])  # Set text color to score_color
        glWindowPos2f(10, self.window_size[1] - 30)  # Position the text
        score_text = f"Score: {self.score}"
        for ch in score_text:
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, ord(ch))

    def Run(self):
        return Process(self)

    def Compile(self):
        with open(inspect.stack()[1].filename, 'r') as file:
            code = file.read()
        with open(os.path.join(os.path.dirname(inspect.stack()[1].filename), 'temp_file.py'), 'w') as file:
            file.write(re.sub('Compile', 'Run', code))
        py_compile.compile(os.path.join(os.path.dirname(inspect.stack()[1].filename), 'temp_file.py'), optimize=2)
        os.remove(os.path.join(os.path.dirname(inspect.stack()[1].filename), 'temp_file.py'))
        shutil.move(f"{os.path.dirname(__file__)}\\__pycache__\\flappy_engine.cpython-311.pyc", f"{os.path.dirname(inspect.stack()[1].filename)}\\__pycache__\\flappy_engine.pyc")
        shutil.move(f"{os.path.dirname(inspect.stack()[1].filename)}\\__pycache__\\temp_file.cpython-311.opt-2.pyc", f"{os.path.dirname(inspect.stack()[1].filename)}\\__pycache__\\{os.path.basename(inspect.stack()[1].filename).replace('.py', '.pyc')}")
        os.rmdir(f"{os.path.dirname(__file__)}\\__pycache__")